@@ -1,6 +1,5 @@
 """Task runner"""
 
-import sys
 from multiprocessing import Process, Queue as MQueue
 from threading import Thread
 import pickle
@@ -10,7 +9,7 @@ import cloudpickle
 
 from .exceptions import InvalidTask, CatchedException
 from .exceptions import TaskFailed, SetupError, DependencyError, UnmetDependency
-from .task import DelayedLoaded
+from .task import Stream, DelayedLoaded
 
 
 # execution result.
@@ -18,7 +17,7 @@ SUCCESS = 0
 FAILURE = 1
 ERROR = 2
 
-class Runner(object):
+class Runner():
     """Task runner
 
     run_all()
@@ -31,19 +30,19 @@ class Runner(object):
 
     """
     def __init__(self, dep_manager, reporter, continue_=False,
-                 always_execute=False, verbosity=0):
+                 always_execute=False, stream=None):
         """
         @param dep_manager: DependencyBase
         @param reporter: reporter object to be used
         @param continue_: (bool) execute all tasks even after a task failure
         @param always_execute: (bool) execute even if up-to-date or ignored
-        @param verbosity: (int) 0,1,2 see Task.execute
+        @param stream: (task.Stream) global verbosity
         """
         self.dep_manager = dep_manager
         self.reporter = reporter
         self.continue_ = continue_
         self.always_execute = always_execute
-        self.verbosity = verbosity
+        self.stream = stream if stream else Stream(0)
 
         self.teardown_list = [] # list of tasks to be teardown
         self.final_result = SUCCESS # until something fails
@@ -101,7 +100,7 @@ class Runner(object):
 
         Tasks should be executed if they are not up-to-date.
 
-        Tasks that cointains setup-tasks must be selected twice,
+        Tasks that contains setup-tasks must be selected twice,
         so it gives chance for dependency tasks to be executed after
         checking it is not up-to-date.
         """
@@ -112,6 +111,9 @@ class Runner(object):
 
             self.reporter.get_status(task)
 
+            # overwrite with effective verbosity
+            task.overwrite_verbosity(self.stream)
+
             # check if task should be ignored (user controlled)
             if node.ignored_deps or self.dep_manager.status_is_ignore(task):
                 node.run_status = 'ignore'
@@ -171,7 +173,7 @@ class Runner(object):
 
         # finally execute it!
         self.reporter.execute_task(task)
-        return task.execute(sys.stdout, sys.stderr, self.verbosity)
+        return task.execute(self.stream)
 
 
     def process_task_result(self, node, catched_excp):
@@ -179,13 +181,20 @@ class Runner(object):
         task = node.task
         # save execution successful
         if catched_excp is None:
-            node.run_status = "successful"
             task.save_extra_values()
-            self.dep_manager.save_success(task)
-            self.reporter.add_success(task)
+            try:
+                self.dep_manager.save_success(task)
+            except FileNotFoundError as exception:
+                msg = ("ERROR: Task '{}' saving success: " \
+                       "Dependent file '{}' does not exist".format(
+                           task.name, exception.filename))
+                catched_excp = DependencyError(msg)
+            else:
+                node.run_status = "successful"
+                self.reporter.add_success(task)
+                return
         # task error
-        else:
-            self._handle_task_error(node, catched_excp)
+        self._handle_task_error(node, catched_excp)
 
 
     def run_tasks(self, task_dispatcher):
@@ -217,8 +226,7 @@ class Runner(object):
         """run teardown from all tasks"""
         for task in reversed(self.teardown_list):
             self.reporter.teardown_task(task)
-            catched = task.execute_teardown(sys.stdout, sys.stderr,
-                                            self.verbosity)
+            catched = task.execute_teardown(self.stream)
             if catched:
                 msg = "ERROR: task '%s' teardown action" % task.name
                 error = SetupError(msg, catched)
@@ -242,7 +250,8 @@ class Runner(object):
         """
         try:
             if hasattr(self.reporter, 'initialize'):
-                self.reporter.initialize(task_dispatcher.tasks)
+                self.reporter.initialize(task_dispatcher.tasks,
+                                         task_dispatcher.selected_tasks)
             self.run_tasks(task_dispatcher)
         except InvalidTask as exception:
             self.reporter.runtime_error(str(exception))
@@ -331,9 +340,9 @@ class MRunner(Runner):
 
     def __init__(self, dep_manager, reporter,
                  continue_=False, always_execute=False,
-                 verbosity=0, num_process=1):
+                 stream=None, num_process=1):
         Runner.__init__(self, dep_manager, reporter, continue_=continue_,
-                        always_execute=always_execute, verbosity=verbosity)
+                        always_execute=always_execute, stream=stream)
         self.num_process = num_process
 
         self.free_proc = 0   # number of free process
@@ -425,16 +434,12 @@ class MRunner(Runner):
 
     def _process_result(self, node, task, result):
         """process result received from sub-process"""
-        if 'failure' in result:
-            catched_excp = result['failure']
-        else:
-            # success set values taken from subprocess result
-            catched_excp = None
-            task.update_from_pickle(result['task'])
-            for action, output in zip(task.actions, result['out']):
-                action.out = output
-            for action, output in zip(task.actions, result['err']):
-                action.err = output
+        catched_excp = result.get('failure')
+        task.update_from_pickle(result['task'])
+        for action, output in zip(task.actions, result['out']):
+            action.out = output
+        for action, output in zip(task.actions, result['err']):
+            action.err = output
         self.process_task_result(node, catched_excp)
 
 
@@ -532,14 +537,13 @@ class MRunner(Runner):
                     continue # pragma: no cover
 
                 result = {'name': task.name}
-                t_result = self.execute_task(task)
+                task_failure = self.execute_task(task)
+                if task_failure:
+                    result['failure'] = task_failure
+                result['task'] = task.pickle_safe_dict()
+                result['out'] = [action.out for action in task.actions]
+                result['err'] = [action.err for action in task.actions]
 
-                if t_result is None:
-                    result['task'] = task.pickle_safe_dict()
-                    result['out'] = [a.out for a in task.actions]
-                    result['err'] = [a.err for a in task.actions]
-                else:
-                    result['failure'] = t_result
                 result_q.put(result)
         except (SystemExit, KeyboardInterrupt, Exception) as exception:
             # error, blow-up everything. send exception info to master process
