@@ -1,12 +1,11 @@
 
 """Tasks are the main abstractions managed by doit"""
 
-import types
 import os
 import sys
 import inspect
 from collections import OrderedDict
-from functools import partial
+from collections.abc import Callable
 from pathlib import PurePath
 
 from .cmdparse import CmdOption, TaskParse
@@ -36,7 +35,7 @@ class DelayedLoader(object):
                    tasks and targets specified on command line
     :ivar target_regex: (str) regex for all targets that this loader tasks
                         will create
-    :ivar created: (bool) wheather this creator was already executed or not
+    :ivar created: (bool) whether this creator was already executed or not
     """
     def __init__(self, creator, executed=None, target_regex=None, creates=None):
         self.creator = creator
@@ -46,12 +45,58 @@ class DelayedLoader(object):
         self.target_regex = target_regex
         self.creates = creates[:] if creates else []
         self.regex_groups = OrderedDict()  # task_name:RegexGroup
+        self.kwargs = None # task creator kwargs
 
 
 # used to indicate that a task had DelayedLoader but was already created
 DelayedLoaded = False
 
 
+class Stream():
+    """Control task output stream verbosity
+
+    :ivar verbosity: 0,1,2 see Task.execute
+
+    Priority is given by:
+    1) command line -> force_global=True
+    2) task value
+    3) other config (INI, DOIT_CONFIG)
+    """
+
+    def __init__(self, verbosity, force_global=False):
+        self.force_global = force_global
+        if verbosity is not None:
+            self.verbosity = verbosity
+        else:
+            self.verbosity = Task.DEFAULT_VERBOSITY
+            self.force_global = False
+
+    def effective_verbosity(self, task_verbosity):
+        """return effective verbosity used on task"""
+        if self.force_global:
+            return self.verbosity
+        elif task_verbosity is not None:
+            return task_verbosity
+        else:
+            return self.verbosity
+
+    @staticmethod
+    def _get_out_err(verbosity):
+        """return tuple (out, err) streams to be used
+
+        Replace stream with None if stream should be captured
+
+        :param verbosity: (int)
+        """
+        if verbosity == 0:
+            return (None, None)
+        elif verbosity == 1:
+            return (None, sys.stderr)
+        else:
+            return (sys.stdout, sys.stderr)
+
+
+
 class Task(object):
     """Task
 
@@ -73,12 +118,13 @@ class Task(object):
                            task values. Always executed on main process.
                            To be used by `uptodate` implementations.
     @ivar setup_tasks (list - string): references to task-names
-    @ivar is_subtask: (bool) indicate this task is a subtask
+    @ivar subtask_of: (string) indicate this task is a subtask of task name
     @ivar has_subtask: (bool) indicate this task has subtasks
     @ivar result: (str) last action "result". used to check task-result-dep
     @ivar values: (dict) values saved by task that might be used by other tasks
     @ivar getargs: (dict) values from other tasks
     @ivar doc: (string) task documentation
+    @ivar meta: (dict) extra info from user/plugin not directly used by doit
 
     @ivar options: (dict) calculated params values (from getargs and taskopt)
     @ivar taskopt: (cmdparse.CmdParse)
@@ -108,18 +154,19 @@ class Task(object):
                   'pos_arg': (string_types, (None,)),
                   'verbosity': ((), (None, 0, 1, 2,)),
                   'getargs': ((dict,), ()),
-                  'title': ((types.FunctionType,), (None,)),
+                  'title': ((Callable,), (None,)),
                   'watch': ((list, tuple), ()),
-    }
+                  'meta': ((dict,), (None,))
+                  }
 
 
     def __init__(self, name, actions, file_dep=(), targets=(),
                  task_dep=(), uptodate=(),
                  calc_dep=(), setup=(), clean=(), teardown=(),
-                 is_subtask=False, has_subtask=False,
+                 subtask_of=None, has_subtask=False,
                  doc=None, params=(), pos_arg=None,
                  verbosity=None, title=None, getargs=None,
-                 watch=(), loader=None):
+                 watch=(), meta=None, loader=None):
         """sanity checks and initialization
 
         @param params: (list of dict for parameters) see cmdparse.CmdOption
@@ -145,12 +192,14 @@ class Task(object):
         self.check_attr(name, 'getargs', getargs, self.valid_attr['getargs'])
         self.check_attr(name, 'title', title, self.valid_attr['title'])
         self.check_attr(name, 'watch', watch, self.valid_attr['watch'])
+        self.check_attr(name, 'meta', meta, self.valid_attr['meta'])
 
         if '=' in name:
             msg = "Task '{}': name must not use the char '=' (equal sign)."
             raise InvalidTask(msg.format(name))
         self.name = name
         self.params = params # save just for use on command `info`
+        self.creator_params = [] # add through task_params decorator
         self.options = None
         self.pos_arg = pos_arg
         self.pos_arg_val = None # to be set when parsing command line
@@ -180,12 +229,13 @@ class Task(object):
         self.uptodate = self._init_uptodate(uptodate)
 
         self.targets = self._init_targets(targets)
-        self.is_subtask = is_subtask
+        self.subtask_of = subtask_of
         self.has_subtask = has_subtask
         self.result = None
         self.values = {}
         self.verbosity = verbosity
         self.custom_title = title
+        self.cfg_values = None
 
         # clean
         if clean is True:
@@ -193,11 +243,14 @@ class Task(object):
             self.clean_actions = ()
         else:
             self._remove_targets = False
-            self.clean_actions = [create_action(a, self) for a in clean]
+            self.clean_actions = [create_action(a, self, 'clean') for a in clean]
 
-        self.teardown = [create_action(a, self) for a in teardown]
+        self.teardown = [create_action(a, self, 'teardown') for a in teardown]
         self.doc = self._init_doc(doc)
         self.watch = watch
+        self.meta = meta
+        # just indicate if actions were executed at all
+        self.executed = False
 
 
     def _init_deps(self, file_dep, task_dep, calc_dep):
@@ -312,17 +365,28 @@ class Task(object):
             self._expand_map[dep](self, dep_values)
 
 
-    def init_options(self):
+    def init_options(self, args=None):
         """Put default values on options.
 
-        This will only be used, if params options were not passed
-        on the command line.
+        This function will only initialize task options once. If provided the args
+        parameter will be parsed for command line arguments intended for this task.
+
+        Return value: unparsed command line task arguments or None.
         """
         if self.options is None:
-            taskcmd = TaskParse([CmdOption(opt) for opt in self.params])
-            # ignore positional parameters
-            self.options = taskcmd.parse('')[0]
+            self.options = {}
+            taskcmd = TaskParse([CmdOption(opt) for opt in (list(self.params) + self.creator_params)])
+            if self.cfg_values is not None:
+                taskcmd.overwrite_defaults(self.cfg_values)
 
+            if args is None or len(args) == 0:
+                # ignore positional parameters
+                self.options.update(taskcmd.parse('')[0])
+            elif len(args) > 0:
+                parsed_options, args = taskcmd.parse(args)
+                self.options.update(parsed_options)
+
+            return args
 
     def _init_getargs(self):
         """task getargs attribute define implicit task dependencies"""
@@ -361,12 +425,12 @@ class Task(object):
         @param valid (list): of valid types/value accepted
         @raises InvalidTask if invalid input
         """
-        if type(value) in valid[0]:
+        if isinstance(value, valid[0]):
             return
         if value in valid[1]:
             return
 
-        # input value didnt match any valid type/value, raise execption
+        # input value didnt match any valid type/value, raise exception
         msg = "Task '%s' attribute '%s' must be " % (task, attr)
         accept = ", ".join([getattr(v, '__name__', str(v)) for v in
                             (valid[0] + valid[1])])
@@ -379,7 +443,7 @@ class Task(object):
         """lazy creation of action instances"""
         if self._action_instances is None:
             self._action_instances = [
-                create_action(a, self) for a in self._actions]
+                create_action(a, self, 'actions') for a in self._actions]
         return self._action_instances
 
 
@@ -388,26 +452,16 @@ class Task(object):
         for value_saver in self.value_savers:
             self.values.update(value_saver())
 
+    def overwrite_verbosity(self, stream):
+        self.verbosity = stream.effective_verbosity(self.verbosity)
 
-    def _get_out_err(self, out, err, verbosity):
-        """select verbosity to be used"""
-        priority = (verbosity, # use command line option
-                    self.verbosity, # or task default from dodo file
-                    self.DEFAULT_VERBOSITY) # or global default
-        use_verbosity = [v for v in  priority if v is not None][0]
-
-        out_err = [(None, None), # 0
-                   (None, err),  # 1
-                   (out, err)]   # 2
-        return out_err[use_verbosity]
-
-
-    def execute(self, out=None, err=None, verbosity=None):
+    def execute(self, stream):
         """Executes the task.
         @return failure: see CmdAction.execute
         """
+        self.executed = True
         self.init_options()
-        task_stdout, task_stderr = self._get_out_err(out, err, verbosity)
+        task_stdout, task_stderr = stream._get_out_err(self.verbosity)
         for action in self.actions:
             action_return = action.execute(task_stdout, task_stderr)
             if isinstance(action_return, CatchedException):
@@ -416,11 +470,11 @@ class Task(object):
             self.values.update(action.values)
 
 
-    def execute_teardown(self, out=None, err=None, verbosity=None):
+    def execute_teardown(self, stream):
         """Executes task's teardown
         @return failure: see CmdAction.execute
         """
-        task_stdout, task_stderr = self._get_out_err(out, err, verbosity)
+        task_stdout, task_stderr = stream._get_out_err(self.verbosity)
         for action in self.teardown:
             action_return = action.execute(task_stdout, task_stderr)
             if isinstance(action_return, CatchedException):
@@ -444,12 +498,14 @@ class Task(object):
                 outstream.write(msg % (self.name, action))
 
                 # add extra arguments used by clean actions
+                execute_on_dryrun = False
                 if isinstance(action, PythonAction):
                     action_sig = inspect.signature(action.py_callable)
                     if 'dryrun' in action_sig.parameters:
+                        execute_on_dryrun = True
                         action.kwargs['dryrun'] = dryrun
 
-                if not dryrun:
+                if (not dryrun) or execute_on_dryrun:
                     result = action.execute(out=outstream)
                     if isinstance(result, CatchedException):
                         sys.stderr.write(str(result))
@@ -535,31 +591,22 @@ def dict_to_task(task_dict):
 
 def clean_targets(task, dryrun):
     """remove all targets from a task"""
-    files = [path for path in task.targets if os.path.isfile(path)]
-    dirs = [path for path in task.targets if os.path.isdir(path)]
-
-    # remove all files
-    for file_ in files:
-        print("%s - removing file '%s'" % (task.name, file_))
-        if not dryrun:
-            os.remove(file_)
-
-    # remove all directories (if empty)
-    for dir_ in dirs:
-        if os.listdir(dir_):
-            msg = "%s - cannot remove (it is not empty) '%s'"
-            print(msg % (task.name, dir_))
-        else:
-            msg = "%s - removing dir '%s'"
-            print(msg % (task.name, dir_))
+    for target in sorted(task.targets, reverse=True):
+        if os.path.isfile(target):
+            print("%s - removing file '%s'" % (task.name, target))
             if not dryrun:
-                os.rmdir(dir_)
+                os.remove(target)
+        elif os.path.isdir(target):
+            if os.listdir(target):
+                msg = "%s - cannot remove (it is not empty) '%s'"
+                print(msg % (task.name, target))
+            else:
+                msg = "%s - removing dir '%s'"
+                print(msg % (task.name, target))
+                if not dryrun:
+                    os.rmdir(target)
 
 
-def _return_param(val):
-    '''just return passed parameter - make a callable from any value'''
-    return val
-
 # uptodate
 class result_dep(UptodateCalculator):
     """check if result of the given task was modified
@@ -596,15 +643,23 @@ class result_dep(UptodateCalculator):
                 sub_tasks[sub] = self.get_val(sub, 'result:')
         return sub_tasks
 
-    def __call__(self, task, values):
-        """return True if result is the same as last run"""
-        dep_task = self.tasks_dict[self.dep_name]
+    def _get_dep_result(self, dep_task):
         if not dep_task.has_subtask:
             dep_result = self._result_single()
         else:
             dep_result = self._result_group(dep_task)
-        func = partial(_return_param, {self.result_name: dep_result})
-        task.value_savers.append(func)
+        return dep_result
+
+
+    def __call__(self, task, values):
+        """return True if result is the same as last run"""
+        dep_task = self.tasks_dict[self.dep_name]
+        dep_result = self._get_dep_result(dep_task)
+
+        def result_saver():
+            # get latest value after execution of dependent task
+            return {self.result_name: self._get_dep_result(dep_task)}
+        task.value_savers.append(result_saver)
 
         last_success = values.get(self.result_name)
         if last_success is None:
